#
#   Darknet RegionLoss
#   Copyright EAVISE
#

import logging
import math
import numpy as np
import torch
import torch.nn as nn
from distutils.version import LooseVersion

try:
    import pandas as pd
except ModuleNotFoundError:
    pd = None


__all__ = ['RegionLoss']
log = logging.getLogger(__name__)
torchversion = LooseVersion(torch.__version__)
version120 = LooseVersion("1.2.0")


class RegionLoss(nn.modules.loss._Loss):
    """ Computes region loss from darknet network output and target annotation (yoloV2).

    Args:
        num_classes (int): number of classes to detect
        anchors (list): 2D list representing anchor boxes (see :class:`lightnet.network.Darknet`)
        stride (optional, int): The downsampling factor of the network (input_dimension / output_dimension); Default **32**
        seen (optional, torch.Tensor): How many images the network has already been trained on; Default **0**
        coord_scale (optional, float): weight of bounding box coordinates; Default **1.0**
        noobject_scale (optional, float): weight of regions without target boxes; Default **1.0**
        object_scale (optional, float): weight of regions with target boxes; Default **5.0**
        class_scale (optional, float): weight of categorical predictions; Default **1.0**
        thresh (optional, float): minimum iou between a predicted box and ground truth for them to be considered matching; Default **0.6**
        coord_prefill (optional, int): This parameter controls for how many training samples the network will prefill the target coordinates, biassing the network to predict the center at **.5,.5**; Default **12800**
    """
    def __init__(self, num_classes, anchors, stride=32, seen=0, coord_scale=1.0, noobject_scale=1.0, object_scale=5.0, class_scale=1.0, thresh=0.6, coord_prefill=12800):
        super().__init__()
        self.num_classes = num_classes
        self.stride = stride
        self.num_anchors = len(anchors)
        self.anchor_step = len(anchors[0])
        self.anchors = torch.tensor(anchors, dtype=torch.float, requires_grad=False)
        self.register_buffer('seen', torch.tensor(seen))

        self.coord_scale = coord_scale
        self.noobject_scale = noobject_scale
        self.object_scale = object_scale
        self.class_scale = class_scale
        self.thresh = thresh
        self.coord_prefill = coord_prefill

        self.mse = nn.MSELoss(reduction='sum')
        self.cel = nn.CrossEntropyLoss(reduction='sum')

        self.loss_total = torch.tensor(0.0)
        self.loss_conf = torch.tensor(0.0)
        self.loss_coord = torch.tensor(0.0)
        self.loss_class = torch.tensor(0.0)

    @property
    def values(self):
        """ Return detached sub-losses in a dictionary.

        Note:
            You can access the individual loss values directly as ``object.loss_<name>`` as well. |br|
            This will return the actual loss tensor with its attached computational graph and gives you full freedom for modifying this loss prior to the backward pass.
        """
        return {
            'total': self.loss_total.detach(),
            'conf':  self.loss_conf.detach(),
            'coord': self.loss_coord.detach(),
            'class': self.loss_class.detach(),
        }

    @property
    def loss(self):
        log.deprecated('The "loss" attribute is deprecated in favor for "loss_total"')
        return self.loss_total

    def extra_repr(self):
        repr_str = f'classes={self.num_classes}, stride={self.stride}, threshold={self.thresh}, seen={self.seen.item()}\n'
        repr_str += f'coord_scale={self.coord_scale}, object_scale={self.object_scale}, noobject_scale={self.noobject_scale}, class_scale={self.class_scale}\n'
        repr_str += f'anchors='
        for a in self.anchors:
            repr_str += f'[{a[0]:.5g}, {a[1]:.5g}] '
        return repr_str

    def forward(self, output, target, seen=None):
        """ Compute Region loss.

        Args:
            output (torch.autograd.Variable): Output from the network
            target (brambox annotation dataframe or torch.Tensor): Brambox annotations or tensor containing the annotation targets (see :class:`lightnet.data.BramboxToTensor`)
            seen (int, optional): How many images the network has already been trained on; Default **Add batch_size to previous seen value**

        Note:
            If using a target tensor, it should have the dimensions `[num_batch, num_anno, 5]` and following format per image:

            .. math::

                \\begin{bmatrix}
                    class\\_idx & x\\_center & y\\_center & width & height \\\\
                    class\\_idx & x\\_center & y\\_center & width & height \\\\
                    ... \\\\
                    -1 & 0 & 0 & 0 & 0 \\\\
                    -1 & 0 & 0 & 0 & 0 \\\\
                    ...
                \\end{bmatrix}

            With all coordinates being relative to the image size. |br|
            Since the annotations from all images of a batch should be made of the same length, you can pad them with: `[-1, 0, 0, 0, 0]`.

        Note:
            Besides being easier to work with, brambox dataframes have the added benefit that
            this loss function will also consider the ``ignore`` flag of annotations and ignore detections that match with it.
            This allows you to have annotations that will not influence the loss in any way,
            as opposed to having them removed and counting them as false detections.
        """
        # Parameters
        nB = output.data.size(0)
        nA = self.num_anchors
        nC = self.num_classes
        nH = output.data.size(2)
        nW = output.data.size(3)
        nPixels = nH * nW
        device = output.device
        if seen is not None:
            self.seen = torch.tensor(seen)
        elif self.training:
            self.seen += nB

        # Get x,y,w,h,conf,cls
        output = output.view(nB, nA, -1, nPixels)
        coord = torch.zeros_like(output[:, :, :4])
        coord[:, :, :2] = output[:, :, :2].sigmoid()    # tx,ty
        coord[:, :, 2:4] = output[:, :, 2:4]            # tw,th
        conf = output[:, :, 4].sigmoid()
        if nC > 1:
            cls = output[:, :, 5:].contiguous().view(nB*nA, nC, nPixels).transpose(1, 2).contiguous().view(-1, nC)

        # Create prediction boxes
        pred_boxes = torch.FloatTensor(nB*nA*nPixels, 4)
        lin_x = torch.linspace(0, nW-1, nW).repeat(nH, 1).view(nPixels).to(device)
        lin_y = torch.linspace(0, nH-1, nH).view(nH, 1).repeat(1, nW).view(nPixels).to(device)
        anchor_w = self.anchors[:, 0].contiguous().view(nA, 1).to(device)
        anchor_h = self.anchors[:, 1].contiguous().view(nA, 1).to(device)

        pred_boxes[:, 0] = (coord[:, :, 0].detach() + lin_x).view(-1)
        pred_boxes[:, 1] = (coord[:, :, 1].detach() + lin_y).view(-1)
        pred_boxes[:, 2] = (coord[:, :, 2].detach().exp() * anchor_w).view(-1)
        pred_boxes[:, 3] = (coord[:, :, 3].detach().exp() * anchor_h).view(-1)
        pred_boxes = pred_boxes.cpu()

        # Get target values
        coord_mask, conf_mask, cls_mask, tcoord, tconf, tcls = self.build_targets(pred_boxes, target, nB, nH, nW)
        coord_mask = coord_mask.expand_as(tcoord).to(device).sqrt()
        conf_mask = conf_mask.to(device).sqrt()
        tcoord = tcoord.to(device)
        tconf = tconf.to(device)
        if nC > 1:
            tcls = tcls[cls_mask].view(-1).long().to(device)
            cls_mask = cls_mask.view(-1, 1).repeat(1, nC).to(device)
            cls = cls[cls_mask].view(-1, nC)

        # Compute losses
        self.loss_coord = self.coord_scale * self.mse(coord*coord_mask, tcoord*coord_mask) / (2 * nB)
        self.loss_conf = self.mse(conf*conf_mask, tconf*conf_mask) / (2 * nB)
        if nC > 1:
            if tcls.numel() > 0:
                self.loss_class = self.class_scale * self.cel(cls, tcls) / nB
            else:
                self.loss_class = torch.tensor(0.0, device=device)
        else:
            self.loss_class = torch.tensor(0.0, device=device)

        self.loss_total = self.loss_coord + self.loss_conf + self.loss_class
        return self.loss_total

    def build_targets(self, pred_boxes, ground_truth, nB, nH, nW):
        """ Compare prediction boxes and targets, convert targets to network output tensors """
        if torch.is_tensor(ground_truth):
            return self.__build_targets_tensor(pred_boxes, ground_truth, nB, nH, nW)
        elif pd is not None and isinstance(ground_truth, pd.DataFrame):
            return self.__build_targets_brambox(pred_boxes, ground_truth, nB, nH, nW)
        else:
            raise TypeError(f'Unkown ground truth format [{type(ground_truth)}]')

    def __build_targets_tensor(self, pred_boxes, ground_truth, nB, nH, nW):
        """ Compare prediction boxes and ground truths, convert ground truths to network output tensors """
        # Parameters
        nT = ground_truth.size(1)
        nA = self.num_anchors
        nAnchors = nA*nH*nW
        nPixels = nH*nW

        # Tensors
        coord_mask = torch.zeros(nB, nA, nH, nW, requires_grad=False)
        conf_mask = torch.ones(nB, nA, nH, nW, requires_grad=False) * self.noobject_scale
        if torchversion >= version120:
            cls_mask = torch.zeros(nB, nA, nH, nW, dtype=torch.bool, requires_grad=False)
        else:
            cls_mask = torch.zeros(nB, nA, nH, nW, requires_grad=False).byte()
        tcoord = torch.zeros(nB, nA, 4, nH, nW, requires_grad=False)
        tconf = torch.zeros(nB, nA, nH, nW, requires_grad=False)
        tcls = torch.zeros(nB, nA, nH, nW, requires_grad=False)

        if self.training and self.seen < self.coord_prefill:
            coord_mask.fill_(math.sqrt(.01 / self.coord_scale))
            if self.anchor_step == 4:
                tcoord[:, :, 0] = self.anchors[:, 2].contiguous().view(1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
                tcoord[:, :, 1] = self.anchors[:, 3].contiguous().view(1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
            else:
                tcoord[:, :, 0].fill_(0.5)
                tcoord[:, :, 1].fill_(0.5)

        # Anchors
        if self.anchor_step == 4:
            anchors = self.anchors.clone()
            anchors[:, :2] = 0
        else:
            anchors = torch.cat([torch.zeros_like(self.anchors), self.anchors], 1)

        # Loop over GT
        for b in range(nB):
            gt = ground_truth[b][(ground_truth[b, :, 0] >= 0)[:, None].expand_as(ground_truth[b])].view(-1, 5)
            if gt.numel() == 0:     # No gt for this image
                continue

            # Build up tensors
            cur_pred_boxes = pred_boxes[b*nAnchors:(b+1)*nAnchors]
            gt = gt[:, 1:]
            gt[:, ::2] *= nW
            gt[:, 1::2] *= nH

            # Set confidence mask of matching detections to 0
            iou_gt_pred = bbox_ious(gt, cur_pred_boxes)
            mask = (iou_gt_pred > self.thresh).sum(0) >= 1
            conf_mask[b][mask.view_as(conf_mask[b])] = 0

            # Find best anchor for each gt
            iou_gt_anchors = bbox_wh_ious(gt, anchors)
            _, best_anchors = iou_gt_anchors.max(1)

            # Set masks and target values for each gt
            nGT = gt.shape[0]
            gi = gt[:, 0].clamp(0, nW-1).long()
            gj = gt[:, 1].clamp(0, nH-1).long()

            conf_mask[b, best_anchors, gj, gi] = self.object_scale
            tconf[b, best_anchors, gj, gi] = iou_gt_pred.view(nGT, nA, nH, nW)[torch.arange(nGT), best_anchors, gj, gi]
            coord_mask[b, best_anchors, gj, gi] = 2 - (gt[:, 2] * gt[:, 3]) / nPixels
            tcoord[b, best_anchors, 0, gj, gi] = gt[:, 0] - gi.float()
            tcoord[b, best_anchors, 1, gj, gi] = gt[:, 1] - gj.float()
            tcoord[b, best_anchors, 2, gj, gi] = (gt[:, 2] / self.anchors[best_anchors, 0]).log()
            tcoord[b, best_anchors, 3, gj, gi] = (gt[:, 3] / self.anchors[best_anchors, 1]).log()
            cls_mask[b, best_anchors, gj, gi] = 1
            tcls[b, best_anchors, gj, gi] = ground_truth[b, torch.arange(nGT), 0]

        return (
            coord_mask.view(nB, nA, 1, nPixels),
            conf_mask.view(nB, nA, nPixels),
            cls_mask.view(nB, nA, nPixels),
            tcoord.view(nB, nA, 4, nPixels),
            tconf.view(nB, nA, nPixels),
            tcls.view(nB, nA, nPixels)
        )

    def __build_targets_brambox(self, pred_boxes, ground_truth, nB, nH, nW):
        """ Compare prediction boxes and ground truths, convert ground truths to network output tensors """
        # Parameters
        nA = self.num_anchors
        nAnchors = nA*nH*nW
        nPixels = nH*nW

        # Tensors
        coord_mask = torch.zeros(nB, nA, nH, nW, requires_grad=False)
        conf_mask = torch.ones(nB, nA, nH, nW, requires_grad=False) * self.noobject_scale
        if torchversion >= version120:
            cls_mask = torch.zeros(nB, nA, nH, nW, dtype=torch.bool, requires_grad=False)
        else:
            cls_mask = torch.zeros(nB, nA, nH, nW, requires_grad=False).byte()
        tcoord = torch.zeros(nB, nA, 4, nH, nW, requires_grad=False)
        tconf = torch.zeros(nB, nA, nH, nW, requires_grad=False)
        tcls = torch.zeros(nB, nA, nH, nW, requires_grad=False)

        if self.training and self.seen < self.coord_prefill:
            coord_mask.fill_(math.sqrt(.01 / self.coord_scale))
            if self.anchor_step == 4:
                tcoord[:, :, 0] = self.anchors[:, 2].contiguous().view(1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
                tcoord[:, :, 1] = self.anchors[:, 3].contiguous().view(1, nA, 1, 1).repeat(nB, 1, 1, nPixels)
            else:
                tcoord[:, :, 0].fill_(0.5)
                tcoord[:, :, 1].fill_(0.5)

        # Anchors
        if self.anchor_step == 4:
            anchors = self.anchors.clone()
            anchors[:, :2] = 0
        else:
            anchors = torch.cat([torch.zeros_like(self.anchors), self.anchors], 1)

        # Loop over GT
        for b, gt_filtered in ground_truth.groupby('batch_number', sort=False):
            cur_pred_boxes = pred_boxes[b*nAnchors:(b+1)*nAnchors]

            # Create ground_truth tensor
            gt = torch.empty((gt_filtered.shape[0], 4), requires_grad=False)
            gt[:, 2] = torch.from_numpy(gt_filtered.width.values).float() / self.stride
            gt[:, 3] = torch.from_numpy(gt_filtered.height.values).float() / self.stride
            gt[:, 0] = torch.from_numpy(gt_filtered.x_top_left.values).float() / self.stride + (gt[:, 2] / 2)
            gt[:, 1] = torch.from_numpy(gt_filtered.y_top_left.values).float() / self.stride + (gt[:, 3] / 2)

            # Set confidence mask of matching detections to 0
            iou_gt_pred = bbox_ious(gt, cur_pred_boxes)
            mask = (iou_gt_pred > self.thresh).sum(0) >= 1
            conf_mask[b][mask.view_as(conf_mask[b])] = 0

            # Find best anchor for each gt
            iou_gt_anchors = bbox_wh_ious(gt, anchors)
            _, best_anchors = iou_gt_anchors.max(1)

            # Set masks and target values for each gt
            nGT = gt.shape[0]
            gi = gt[:, 0].clamp(0, nW-1).long()
            gj = gt[:, 1].clamp(0, nH-1).long()

            conf_mask[b, best_anchors, gj, gi] = self.object_scale
            tconf[b, best_anchors, gj, gi] = iou_gt_pred.view(nGT, nA, nH, nW)[torch.arange(nGT), best_anchors, gj, gi]
            coord_mask[b, best_anchors, gj, gi] = 2 - (gt[:, 2] * gt[:, 3]) / nPixels
            tcoord[b, best_anchors, 0, gj, gi] = gt[:, 0] - gi.float()
            tcoord[b, best_anchors, 1, gj, gi] = gt[:, 1] - gj.float()
            tcoord[b, best_anchors, 2, gj, gi] = (gt[:, 2] / self.anchors[best_anchors, 0]).log()
            tcoord[b, best_anchors, 3, gj, gi] = (gt[:, 3] / self.anchors[best_anchors, 1]).log()
            cls_mask[b, best_anchors, gj, gi] = 1
            tcls[b, best_anchors, gj, gi] = torch.from_numpy(gt_filtered.class_id.values).float()

            # Set masks of ignored to zero
            if gt_filtered.ignore.any():
                if torchversion >= version120:
                    ignore_mask = torch.from_numpy(gt_filtered.ignore.values)
                else:
                    ignore_mask = torch.from_numpy(gt_filtered.ignore.values.astype(np.uint8))
                gi = gi[ignore_mask]
                gj = gj[ignore_mask]
                best_anchors = best_anchors[ignore_mask]

                conf_mask[b, best_anchors, gj, gi] = 0
                coord_mask[b, best_anchors, gj, gi] = 0
                cls_mask[b, best_anchors, gj, gi] = 0

        return (
            coord_mask.view(nB, nA, 1, nPixels),
            conf_mask.view(nB, nA, nPixels),
            cls_mask.view(nB, nA, nPixels),
            tcoord.view(nB, nA, 4, nPixels),
            tconf.view(nB, nA, nPixels),
            tcls.view(nB, nA, nPixels)
        )


def bbox_ious(boxes1, boxes2):
    """ Compute IOU between all boxes from ``boxes1`` with all boxes from ``boxes2``.

    Args:
        boxes1 (torch.Tensor): List of bounding boxes
        boxes2 (torch.Tensor): List of bounding boxes

    Returns:
        torch.Tensor[len(boxes1) X len(boxes2)]: IOU values

    Note:
        Tensor format: [[xc, yc, w, h],...]
    """
    b1x1, b1y1 = (boxes1[:, :2] - (boxes1[:, 2:4] / 2)).split(1, 1)
    b1x2, b1y2 = (boxes1[:, :2] + (boxes1[:, 2:4] / 2)).split(1, 1)
    b2x1, b2y1 = (boxes2[:, :2] - (boxes2[:, 2:4] / 2)).split(1, 1)
    b2x2, b2y2 = (boxes2[:, :2] + (boxes2[:, 2:4] / 2)).split(1, 1)

    dx = (b1x2.min(b2x2.t()) - b1x1.max(b2x1.t())).clamp(min=0)
    dy = (b1y2.min(b2y2.t()) - b1y1.max(b2y1.t())).clamp(min=0)
    intersections = dx * dy

    areas1 = (b1x2 - b1x1) * (b1y2 - b1y1)
    areas2 = (b2x2 - b2x1) * (b2y2 - b2y1)
    unions = (areas1 + areas2.t()) - intersections

    return intersections / unions


def bbox_wh_ious(boxes1, boxes2):
    """ Shorter version of :func:`lightnet.network.loss._regionloss.bbox_ious`
    for when we are only interested in W/H of the bounding boxes and not X/Y.

    Args:
        boxes1 (torch.Tensor): List of bounding boxes
        boxes2 (torch.Tensor): List of bounding boxes

    Returns:
        torch.Tensor[len(boxes1) X len(boxes2)]: IOU values when discarding X/Y offsets (aka. as if they were zero)

    Note:
        Tensor format: [[xc, yc, w, h],...]
    """
    b1w = boxes1[:, 2].unsqueeze(1)
    b1h = boxes1[:, 3].unsqueeze(1)
    b2w = boxes2[:, 2]
    b2h = boxes2[:, 3]

    intersections = b1w.min(b2w) * b1h.min(b2h)
    unions = (b1w * b1h) + (b2w * b2h) - intersections

    return intersections / unions



#
#   Mobile YOLOv2 model
#   Copyright EAVISE
#

import functools
from collections import OrderedDict, Iterable
import torch
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['MobileYoloV2Upsample']


class MobileYoloV2Upsample(lnn.module.Darknet):
    """ MobileYolo v2 implementation with an upsampling layer instead of a reorg layer. |br|
    TODO : add citation

    This is a variant of :class:`~lightnet.models.MobileYoloV2` where we removed the reorg layer
    and instead upsample the other branch before concatenation.
    This results in a two times bigger output feature map, but only has a limited number of extra computations,
    as there are only two convolutions after the concat operation.

    Note:
        Because the output feature map is two times bigger, we made the anchors two times bigger as well.

    Args:
        num_classes (Number, optional): Number of classes; Default **20**
        input_channels (Number, optional): Number of input channels; Default **3**
        anchors (list, optional): 2D list with anchor values; Default **Yolo v2 anchors**

    Attributes:
        self.stride: Subsampling factor of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
        self.remap_mobile_darknet19: Remapping rules for weights from the :class:`~lightnet.models.MobileDarknet19` model.
    """
    stride = 16
    inner_stride = 32
    remap_mobile_darknet19 = [
        (r'^layers.0.([1-9]_)',     r'layers.0.\1'),    # layers 1-9
        (r'^layers.0.(1[0-3]_)',    r'layers.0.\1'),    # layers 10-13
        (r'^layers.0.(\d{2}_)',     r'layers.1.\1'),    # remaining layers (14-19)
    ]

    def __init__(self, num_classes, input_channels=3, anchors=[(2.6442, 3.4629), (6.3855, 8.01888), (10.11174, 16.19784), (18.94224, 9.68106), (22.4728, 20.0142)]):
        super().__init__()
        if not isinstance(anchors, Iterable) and not isinstance(anchors[0], Iterable):
            raise TypeError('Anchors need to be a 2D list of numbers')

        # Parameters
        self.num_classes = num_classes
        self.input_channels = input_channels
        self.anchors = anchors

        # Network
        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01
        layer_list = [
            # Sequence 0 : input = image tensor
            OrderedDict([
                ('1_convbatch',     lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 2, 1, relu=relu, momentum=momentum)),
                ('2_convdw',        lnn.layer.Conv2dDepthWise(32, 64, 3, 2, 1, relu=relu, momentum=momentum)),
                ('3_convdw',        lnn.layer.Conv2dDepthWise(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ('4_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                ('5_convdw',        lnn.layer.Conv2dDepthWise(64, 128, 3, 2, 1, relu=relu, momentum=momentum)),
                ('6_convdw',        lnn.layer.Conv2dDepthWise(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ('7_convbatch',     lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                ('8_convdw',        lnn.layer.Conv2dDepthWise(128, 256, 3, 2, 1, relu=relu, momentum=momentum)),
                ('9_convdw',        lnn.layer.Conv2dDepthWise(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('10_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                ('11_convdw',       lnn.layer.Conv2dDepthWise(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('12_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                ('13_convdw',       lnn.layer.Conv2dDepthWise(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
            ]),

            # Sequence 1 : input = sequence0
            OrderedDict([
                ('14_max',          nn.MaxPool2d(2, 2)),
                ('15_convdw',       lnn.layer.Conv2dDepthWise(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('16_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                ('17_convdw',       lnn.layer.Conv2dDepthWise(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('18_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                ('19_convdw',       lnn.layer.Conv2dDepthWise(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('20_convdw',       lnn.layer.Conv2dDepthWise(1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('21_convdw',       lnn.layer.Conv2dDepthWise(1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('22_upsample',     torch.nn.Upsample(scale_factor=2, mode='nearest'))
            ]),

            # Sequence 2 : input = sequence0
            OrderedDict([
                ('23_convbatch',    lnn.layer.Conv2dBatchReLU(512, 4*64, 1, 1, 0, relu=relu, momentum=momentum)),
            ]),

            # Sequence 3 : input = sequence2 + sequence1
            OrderedDict([
                ('24_convbatch',    lnn.layer.Conv2dBatchReLU((4*64)+1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('25_conv',         nn.Conv2d(1024, len(self.anchors)*(5+self.num_classes), 1, 1, 0)),
            ])
        ]
        self.layers = nn.ModuleList([nn.Sequential(layer_dict) for layer_dict in layer_list])

    def forward(self, x):
        out0 = self.layers[0](x)
        out1 = self.layers[1](out0)
        out2 = self.layers[2](out0)
        out = self.layers[3](torch.cat((out2, out1), 1))

        return out



#
#   Dependency list related functions
#   Copyright EAVISE
#

from ._dependency import *
from ._tree import traverse_depth_first, traverse_breadth_first
from ._type import NodeType, add_nodetype_operation



#
#   Darknet RegionLoss
#   Copyright EAVISE
#
import logging
import math
import numpy as np
import torch
import torch.nn as nn

try:
    import pandas as pd
except ModuleNotFoundError:
    pd = None

__all__ = ['CornerLoss']
log = logging.getLogger(__name__)


class CornerLoss(nn.modules.loss._Loss):
    """ Computes cornerloss loss from the cornernet network output and target annotation.

    .. admonition:: Experimental

       This loss implementation is still in development
       and might not be yielding the same results as the official implementation.
       Use at your own risk!

    Args:
        stride (int, optional): The downsampling factor of the network (input_dimension / output_dimension); Default **4**
        gaussian_iou (float, optional): Minimal box IoU for corner penalty reduction (set to zero to disable gaussian penalty reduction); Default **0.3**
        heatmap_scale (float, optional): Scale factor for the heatmap corner loss; Default **1**
        pull_scale (float, optional): Scale factor for the pull part of the embedding loss (pull similar classes towards each other); Default **1**
        push_scale (float, optional): Scale factor for the push part of the embedding loss (push different classes from each other); Default **1**
        offset_scale (float, optional): Scale factor for the corner offset loss; Default **1**
        inter_scale (float, optional): Scale factor for the loss computed on the intermediate feature map; Default **1**
    """
    def __init__(self, stride=4, gaussian_iou=0.3, heatmap_scale=1, pull_scale=1, push_scale=1, offset_scale=1, inter_scale=1):
        super().__init__()
        log.experimental(f'"{self.__class__.__name__}" is still in development. Use at your own risk!')

        self.stride = stride
        self.gaussian_iou = gaussian_iou
        self.heatmap_scale = heatmap_scale
        self.pull_scale = pull_scale
        self.push_scale = push_scale
        self.offset_scale = offset_scale
        self.inter_scale = inter_scale

        self.eps = 1e-4
        self.l1 = nn.SmoothL1Loss()

        self.loss_total = torch.tensor(0.0)
        self.loss_heatmap = torch.tensor(0.0)
        self.loss_embedding = torch.tensor(0.0)
        self.loss_offset = torch.tensor(0.0)

    @property
    def values(self):
        """ Return various sub-losses values as a dictionary.

        Note:
            You can access the individual loss values directly as ``object.loss_<name>`` as well. |br|
            This will return the actual loss tensor with its attached computational graph and gives you full freedom for modifying this loss prior to the backward pass.
        """
        return {
            'total': self.loss_total.item(),
            'heatmap': self.loss_heatmap.item(),
            'embedding': self.loss_embedding.item(),
            'offset': self.loss_offset.item(),
        }

    @property
    def loss(self):
        log.deprecated('The "loss" attribute is deprecated in favor for "loss_total"')
        return self.loss_total

    def extra_repr(self):
        repr_str = f'stride={self.stride}, gaussian_iou={self.gaussian_iou}\n'
        repr_str += f'heatmap_scale={self.heatmap_scale}, pull_scale={self.pull_scale}, push_scale={self.push_scale}, offset_scale={self.offset_scale}, inter_scale={self.inter_scale}'
        return repr_str

    def forward(self, output, target):
        """ Compute Corner loss.

        Args:
            output (tuple of 2 Tensors): Output from the network :class:`~lightnet.models.Cornernet` (training mode)
            target (brambox annotation dataframe): Brambox annotations
        """
        if isinstance(output, (list, tuple)):
            output, intermediate = output
        else:
            log.error_once(f'Output does not contain intermediate feature map, which is neccesary for training!')
            intermediate = None

        # Parameters
        device = output.device
        nB, nC, nH, nW = output.shape
        nClasses = (nC // 2) - 3

        # Split output
        output = output.view(nB, 2, -1, nH, nW)                                             # BATCH, TLBR, NUM_CLASSES+3, H, W
        out_heatmaps = output[:, :, :-3].sigmoid().clamp(min=1e-4, max=1-1e-4)              # BATCH, TLBR, NUM_CLASSES,   H, W
        out_embeddings = output[:, :, -3].permute(1, 0, 2, 3)                               # TLBR, BATCH,                H, W
        out_offsets = output[:, :, -2:]                                                     # BATCH, TLBR, XY,            H, W

        # Split intermediate
        if intermediate is not None:
            intermediate = intermediate.view(nB, 2, -1, nH, nW)                             # BATCH, TLBR, NUM_CLASSES+3, H, W
            inter_heatmaps = intermediate[:, :, :-3].sigmoid().clamp(min=1e-4, max=1-1e-4)  # BATCH, TLBR, NUM_CLASSES,   H, W
            inter_embeddings = intermediate[:, :, -3].permute(1, 0, 2, 3)                   # TLBR, BATCH,                H, W
            inter_offsets = intermediate[:, :, -2:]                                         # BATCH, TLBR, XY,            H, W

        # Get ground truth tensors
        gt_heatmaps, gt_embeddings, gt_offsets, gt_mask = self.build_targets(target, nB, nClasses, nH, nW)
        gt_heatmaps = gt_heatmaps.to(device)
        gt_embeddings = gt_embeddings.to(device)
        gt_offsets = gt_offsets.to(device)
        gt_mask = gt_mask[:, :, None, ...].expand_as(gt_offsets).to(device)
        nGT = gt_mask.sum().item()

        # Losses
        if intermediate is not None:
            self.loss_heatmap = self.heatmap_scale * (
                self.focal_loss(out_heatmaps, gt_heatmaps)
                + self.inter_scale * self.focal_loss(inter_heatmaps, gt_heatmaps)
            )

            if nGT > 0:
                self.loss_embedding = (
                    self.pushpull_loss(out_embeddings, gt_embeddings)
                    + self.inter_scale * self.pushpull_loss(inter_embeddings, gt_embeddings)
                )

                self.loss_offset = self.offset_scale * (
                    self.l1(out_offsets[gt_mask], gt_offsets[gt_mask])
                    + self.inter_scale * self.l1(inter_offsets[gt_mask], gt_offsets[gt_mask])
                )
            else:
                self.loss_embedding = torch.tensor(0.0)
                self.loss_offset = torch.tensor(0.0)

            self.loss_total = (self.loss_heatmap + self.loss_embedding + self.loss_offset) / (1 + self.inter_scale)
        else:
            self.loss_heatmap = self.heatmap_scale * self.focal_loss(out_heatmaps, gt_heatmaps)
            if nGT > 0:
                self.loss_embedding = self.pushpull_loss(out_embeddings, gt_embeddings)
                self.loss_offset = self.offset_scale * self.l1(out_offsets[gt_mask], gt_offsets[gt_mask])
            else:
                self.loss_embedding = torch.tensor(0.0)
                self.loss_offset = torch.tensor(0.0)

            self.loss_total = self.loss_heatmap + self.loss_embedding + self.loss_offset

        # Loss
        return self.loss_total

    def build_targets(self, ground_truth, nB, nC, nH, nW):
        """ Convert ground truths to network output tensors """
        heatmaps = torch.zeros(nB, 2, nC, nH, nW, requires_grad=False)
        offsets = torch.zeros(nB, 2, 2, nH, nW, requires_grad=False)
        mask = torch.zeros(nB, 2, nH, nW, dtype=torch.bool, requires_grad=False)
        embedding = list()

        for b, gt_batch in ground_truth.groupby('batch_number', sort=False):
            # GT tensors
            class_id = torch.from_numpy(gt_batch.class_id.values).long()
            size = torch.from_numpy(gt_batch[['width', 'height']].values).float() / self.stride
            coords = torch.empty((gt_batch.shape[0], 4), requires_grad=False)
            coords[:, 0:2] = torch.from_numpy(gt_batch[['x_top_left', 'y_top_left']].values).float() / self.stride
            coords[:, 2:4] = coords[:, 0:2] + size
            coords_idx = coords.long()
            coords_idx[:, 0:3:2].clamp_(max=nW-1)
            coords_idx[:, 1:4:2].clamp_(max=nH-1)

            # Heatmaps
            if self.gaussian_iou:
                radii = gaussian_radius(size[:, 0], size[:, 1], self.gaussian_iou)
                coords0 = (coords_idx - radii[:, None]).clamp(min=0)
                coords1 = (coords_idx + radii[:, None] + 1)
                coords1[:, 0:3:2].clamp_(max=nW)
                coords1[:, 1:4:2].clamp_(max=nH)
                for idx, r in enumerate(radii):
                    g = create_gaussian(r, (r+0.5)/3)

                    sxtl = slice(coords0[idx, 0], coords1[idx, 0])
                    gxtl = slice(0, sxtl.stop - sxtl.start)
                    sytl = slice(coords0[idx, 1], coords1[idx, 1])
                    gytl = slice(0, sytl.stop - sytl.start)

                    sxbr = slice(coords0[idx, 2], coords1[idx, 2])
                    gxbr = slice(0, sxbr.stop - sxbr.start)
                    sybr = slice(coords0[idx, 3], coords1[idx, 3])
                    gybr = slice(0, sybr.stop - sybr.start)

                    heatmaps[b, 0, class_id[idx], sytl, sxtl] = torch.max(
                        g[gytl, gxtl],
                        heatmaps[b, 0, class_id[idx], sytl, sxtl]
                    )

                    heatmaps[b, 1, class_id[idx], sybr, sxbr] = torch.max(
                        g[gybr, gxbr],
                        heatmaps[b, 1, class_id[idx], sybr, sxbr]
                    )
            else:
                heatmaps[b, 0, class_id, coords_idx[:, 1], coords_idx[:, 0]] = 1
                heatmaps[b, 1, class_id, coords_idx[:, 3], coords_idx[:, 2]] = 1

            # Mask
            mask[b, 0, coords_idx[:, 1], coords_idx[:, 0]] = True
            mask[b, 1, coords_idx[:, 3], coords_idx[:, 2]] = True

            # Embeddings
            embedding.append((b * nW * nH) + (coords_idx[:, 1:4:2] * nW) + coords_idx[:, 0:3:2])

            # Offsets
            off = (coords - coords_idx).transpose(0, 1).contiguous()
            offsets[b, 0, :, coords_idx[:, 1], coords_idx[:, 0]] = off[0:2]
            offsets[b, 1, :, coords_idx[:, 3], coords_idx[:, 2]] = off[2:4]

        return (
            heatmaps,
            torch.cat(embedding, dim=0),
            offsets,
            mask
        )

    def focal_loss(self, pred, gt):
        p_mask = gt.eq(1)
        n_mask = ~p_mask

        p_pred = pred[p_mask]
        n_pred = pred[n_mask]

        p_loss = (p_pred.log() * (1 - p_pred) ** 2).sum()
        n_loss = ((1 - n_pred).log() * (n_pred ** 2) * ((1 - gt[n_mask]) ** 4)).sum()

        num_pos = p_mask.sum().item()
        if num_pos == 0:
            return -1 * n_loss
        else:
            return -1 * (p_loss + n_loss) / num_pos

    def pushpull_loss(self, pred, gt):
        nGT = gt.shape[0]
        gt = gt.transpose(0, 1)
        tl = torch.take(pred[0], gt[0])
        br = torch.take(pred[1], gt[1])

        pred_mean = (tl + br) / 2
        pull = ((tl - pred_mean) ** 2 + (br - pred_mean) ** 2).sum() / (nGT + self.eps)

        dist = (1 - torch.abs(pred_mean[..., None] - pred_mean[None, ...])).clamp(min=0)
        push = dist.triu(1).sum() / ((nGT - 1) * nGT + self.eps)

        return self.push_scale * push + self.pull_scale * pull


def gaussian_radius(box_width, box_height, iou):
    box_width = box_width.ceil().int()
    box_height = box_height.ceil().int()

    # a1 = 1
    b1 = (box_width + box_height)
    c1 = box_width * box_height * (1 - iou) / (1 + iou)
    sq1 = (b1 ** 2 - 4 * c1).sqrt()
    r1 = (b1 - sq1) / 2

    # a2  = 4
    b2 = 2 * (box_width + box_height)
    c2 = (1 - iou) * box_width * box_height
    sq2 = (b2 ** 2 - 16 * c2).sqrt()
    r2 = (b2 - sq2) / 8

    a3 = 4 * iou
    b3 = -2 * iou * (box_height + box_width)
    c3 = (iou - 1) * box_width * box_height
    sq3 = (b3 ** 2 - 4 * a3 * c3).sqrt()
    r3 = (b3 + sq3) / (2 * a3)

    rmin, _ = torch.min(torch.stack([r1, r2, r3], dim=0), dim=0)
    return rmin.int().clamp(min=0)


def create_gaussian(radius, sigma=1):
    x = torch.arange(-radius, radius+1, dtype=torch.float)[None, ...]
    y = x.clone().view(-1, 1)
    return (-1 * (x**2 + y**2) / (2 * sigma**2)).exp()



#
#   Lightnet utilitary pre-processing operators
#   Copyright EAVISE
#

import logging
import numpy as np
import torch
from ..util import BaseTransform, BaseMultiTransform
from ..._imports import pd

__all__ = ['AnnoTransform', 'BramboxToTensor']
log = logging.getLogger(__name__)


class AnnoTransform(BaseMultiTransform):
    """ MultiTransform that allows to modify annotations.
        Its main use is to be able to transform annotations when creating a single :class:`~lightnet.data.transform.Compose` transformation pipeline for both images and annotations.

        Because the Compose pipeline will only run transformations on the image data, unless it is a multi-transform,
        it is necessary to encapsulate functions for annotaions in a such a class.

        Args:
            fn (callable): Function that is called with the annotation dataframe
    """
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def _tf_pil(self, img):
        return img

    def _tf_cv(self, img):
        return img

    def _tf_torch(self, img):
        return img

    def _tf_anno(self, anno):
        return self.fn(anno)


class BramboxToTensor(BaseTransform):
    """ Converts a list of brambox annotation objects to a tensor.

    .. deprecated:: 2.0.0
        |br| This class is deprectated, because you can use brambox dataframes in the loss functions.

    Args:
        dimension (tuple, optional): Default size of the transformed images, expressed as a (width, height) tuple; Default **None**
        dataset (lightnet.data.Dataset, optional): Dataset that uses this transform; Default **None**
        max_anno (Number, optional): Maximum number of annotations in the list; Default **50**
        class_label_map (list, optional): class label map to convert class names to an index; Default **None**

    Return:
        torch.Tensor: tensor of dimension [max_anno, 5] containing [class_idx,center_x,center_y,width,height] for every detection

    Warning:
        To convert annotations to a torch Tensor, you need to convert the `class_label` to an integer. |br|
        For this purpose, this function will first check if the dataframe has a `class_index` column to use.
        Otherwise, it will convert the strings by mapping them to the index of the `class_label_map` argument.
        If no class_label_map is given, it will then try to convert the class_label to an integer, using `astype(int)`.
        If that fails, it is simply given the number 0.
    """
    def __init__(self, dimension=None, dataset=None, max_anno=50, class_label_map=None):
        super().__init__()
        log.deprecated('This class is deprecated and will be removed in a future version of lightnet. Please use brambox dataframes instead of tensors in the various loss functions.')
        self.dimension = dimension
        self.dataset = dataset
        self.max_anno = max_anno
        self.class_label_map = class_label_map

        if self.dimension is None and self.dataset is None:
            raise ValueError('This transform either requires a dimension or a dataset to infer the dimension')
        if self.class_label_map is None:
            log.warning('No class_label_map given. If there is no class_index column or if the class_labels are not integers, they will be set to zero.')

    def forward(self, data):
        if self.dataset is not None:
            dim = self.dataset.input_dim
        else:
            dim = self.dimension
        return self.apply(data, dim, self.max_anno, self.class_label_map)

    @classmethod
    def apply(cls, data, dimension, max_anno=None, class_label_map=None):
        if not isinstance(data, pd.DataFrame):
            raise TypeError(f'BramboxToTensor only works with brambox annotation dataframes [{type(data)}]')

        anno_np = cls._tf_anno(data, dimension, class_label_map)

        if max_anno is not None:
            anno_len = len(anno_np)
            if anno_len > max_anno:
                raise ValueError(f'More annotations than maximum allowed [{anno_len}/{max_anno}]')

            z_np = np.zeros((max_anno-anno_len, 5), dtype=np.float32)
            z_np[:, 0] = -1

            if anno_len > 0:
                return torch.from_numpy(np.concatenate((anno_np, z_np)))
            else:
                return torch.from_numpy(z_np)
        else:
            return torch.from_numpy(anno_np)

    @staticmethod
    def _tf_anno(anno, dimension, class_label_map):
        net_w, net_h = dimension

        if 'class_index' not in anno.columns:
            if class_label_map is not None:
                cls_idx = anno.class_label.map(dict((l, i) for i, l in enumerate(class_label_map))).values
            else:
                try:
                    cls_idx = anno.class_label.astype(int).values
                except ValueError:
                    cls_idx = np.array([0] * len(anno))
        else:
            cls_idx = anno['class_index'].values

        w = anno.width.values / net_w
        h = anno.height.values / net_h
        cx = anno.x_top_left.values / net_w + (w / 2)
        cy = anno.y_top_left.values / net_h + (h / 2)

        return np.stack([cls_idx, cx, cy, w, h], axis=-1).astype(np.float32)



#
#   Darknet YOLOv2 model with Mobilenet backend
#   Copyright EAVISE
#

from collections import OrderedDict, Iterable
import functools
import torch
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['MobilenetYolo']


class MobilenetYolo(lnn.module.Lightnet):
    """ Yolo v2 implementation with a mobilenet v1 backend :cite:`mobilenet_v1`.

    Args:
        num_classes (Number, optional): Number of classes; Default **20**
        alpha (Number, optional): Number between [0-1] that controls the number of filters of the mobilenet convolutions; Default **1**
        input_channels (Number, optional): Number of input channels; Default **3**
        anchors (list, optional): 2D list with anchor values; Default **Yolo v2 anchors**

    Attributes:
        self.stride: Subsampling factor of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
        self.remap_mobilenet_v1: Remapping rules for weights from the :class:`~lightnet.models.MobileNetV1` model.

    Warning:
        When changing the ``alpha`` value, you are changing the network architecture.
        This means you cannot use weights from this network with a different alpha value.
    """
    stride = 32
    inner_stride = 32
    remap_mobilenet_v1 = [
        (r'^layers.0.([1-9]_)',     r'layers.0.\1'),    # layers 1-9
        (r'^layers.0.(1[0-4]_)',    r'layers.1.\1'),    # layers 10-14
    ]

    def __init__(self, num_classes, alpha=1.0, input_channels=3, anchors=[(1.3221, 1.73145), (3.19275, 4.00944), (5.05587, 8.09892), (9.47112, 4.84053), (11.2364, 10.0071)]):
        super().__init__()
        if not isinstance(anchors, Iterable) and not isinstance(anchors[0], Iterable):
            raise TypeError('Anchors need to be a 2D list of numbers')

        # Parameters
        self.num_classes = num_classes
        self.anchors = anchors

        # Network
        relu = functools.partial(nn.ReLU6, inplace=True)
        momentum = 0.1
        layer_list = [
            # Sequence 0 : input = image tensor
            OrderedDict([
                ('1_convbatch',     lnn.layer.Conv2dBatchReLU(input_channels, int(alpha*32),  3, 2, 1, relu=relu, momentum=momentum)),
                ('2_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*32),  int(alpha*64),  3, 1, 1, relu=relu, momentum=momentum)),
                ('3_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*64),  int(alpha*128), 3, 2, 1, relu=relu, momentum=momentum)),
                ('4_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*128), int(alpha*128), 3, 1, 1, relu=relu, momentum=momentum)),
                ('5_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*128), int(alpha*256), 3, 2, 1, relu=relu, momentum=momentum)),
                ('6_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*256), int(alpha*256), 3, 1, 1, relu=relu, momentum=momentum)),
                ('7_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*256), int(alpha*512), 3, 2, 1, relu=relu, momentum=momentum)),
                ('8_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu, momentum=momentum)),
                ('9_convdw',        lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu, momentum=momentum)),
            ]),

            # Sequence 1 : input = sequence0
            OrderedDict([
                ('10_convdw',       lnn.layer.Conv2dDepthWise(int(alpha*512),  int(alpha*512),  3, 1, 1, relu=relu, momentum=momentum)),
                ('11_convdw',       lnn.layer.Conv2dDepthWise(int(alpha*512),  int(alpha*512),  3, 1, 1, relu=relu, momentum=momentum)),
                ('12_convdw',       lnn.layer.Conv2dDepthWise(int(alpha*512),  int(alpha*512),  3, 1, 1, relu=relu, momentum=momentum)),
                ('13_convdw',       lnn.layer.Conv2dDepthWise(int(alpha*512),  int(alpha*1024), 3, 2, 1, relu=relu, momentum=momentum)),
                ('14_convdw',       lnn.layer.Conv2dDepthWise(int(alpha*1024), int(alpha*1024), 3, 1, 1, relu=relu, momentum=momentum)),
            ]),

            # Sequence 2 : input = sequence0
            OrderedDict([
                ('15_convbatch',    lnn.layer.Conv2dBatchReLU(int(alpha*512), 64, 1, 1, 0, relu=relu, momentum=momentum)),
                ('16_reorg',        lnn.layer.Reorg(2)),
            ]),

            # Sequence 3 : input = sequence2 + sequence1
            OrderedDict([
                ('17_convbatch',    lnn.layer.Conv2dBatchReLU((4*64)+int(alpha*1024), 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('18_conv',         nn.Conv2d(1024, len(self.anchors)*(5+self.num_classes), 1, 1, 0)),
            ])
        ]
        self.layers = nn.ModuleList([nn.Sequential(layer_dict) for layer_dict in layer_list])

    def forward(self, x):
        out0 = self.layers[0](x)
        out1 = self.layers[1](out0)
        out2 = self.layers[2](out0)
        out = self.layers[3](torch.cat((out2, out1), 1))

        return out



#
#   Mobilenet related layers
#   Copyright EAVISE
#

import logging
import torch.nn as nn


__all__ = ['Conv2dDepthWise', 'InvertedBottleneck']
log = logging.getLogger(__name__)


class Conv2dDepthWise(nn.Module):
    """ This layer implements the depthwise separable convolution :cite:`mobilenet_v1`. |br|
    Instead of performing a regular convolution,
    this layer first does a depthwise convolution, followed by a pointwise convolution.
    This reduces the number of computations, while maintaining a similar accuracy.

    .. figure:: /.static/api/conv2ddepthwise.*
       :width: 100%
       :alt: Conv2dDepthWise module design

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int or tuple): Size of the kernel of the convolution
        stride (int or tuple): Stride of the convolution
        padding (int or tuple): padding of the convolution
        momentum (int, optional): momentum of the moving averages of the normalization; Default **0.1**
        relu (class, optional): Which ReLU to use; Default :class:`torch.nn.ReLU`

    Note:
        The bias term in both :class:`~torch.nn.Conv2d` is disabled for this module.

    Example:
        >>> module = ln.network.layer.Conv2dDepthWise(3, 32, 3, 1, 1)
        >>> print(module)
        Conv2dDepthWise(3, 32, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), ReLU(inplace=True))
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 10, 10])
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride, padding, momentum=0.1, relu=lambda: nn.ReLU(inplace=True)):
        super().__init__()

        self.layers = nn.Sequential(
            nn.Conv2d(in_channels, in_channels, kernel_size, stride, padding, groups=in_channels, bias=False),
            nn.BatchNorm2d(in_channels, momentum=momentum),
            relu(),

            nn.Conv2d(in_channels, out_channels, 1, 1, 0, bias=False),
            nn.BatchNorm2d(out_channels, momentum=momentum),
            relu(),
        )

    def __repr__(self):
        s = '{name}({in_channels}, {out_channels}, kernel_size={kernel_size}, stride={stride}, padding={padding}, {relu})'
        return s.format(
            name=self.__class__.__name__,
            in_channels=self.layers[0].in_channels,
            out_channels=self.layers[3].out_channels,
            kernel_size=self.layers[0].kernel_size,
            stride=self.layers[0].stride,
            padding=self.layers[0].padding,
            relu=self.layers[2],
        )

    def forward(self, x):
        x = self.layers(x)
        return x


class InvertedBottleneck(nn.Module):
    """ This is an implementation of the inverted bottleneck layer :cite:`mobilenet_v2`. |br|
    We first use a pointwise convolution to expand the number of channels followed by a depthwise convolution.
    Finally we use another pointwise convolution to reduce the number of channels to the required amount.
    If the number of input channels is equal to the number of output channels, we add the original input to the output,
    in order to more easily propagate the gradient in the network.

    .. figure:: /.static/api/invertedbottleneck.*
       :width: 100%
       :alt: Conv2dDepthWise module design

    Args:
        in_channels (int): Number of input channels
        out_channels (int): Number of output channels
        kernel_size (int or tuple): Size of the kernel of the convolution
        stride (int or tuple): Stride of the convolution
        expansion (int): Expansion factor for the number of channels in the depthwise convolution
        momentum (int, optional): momentum of the moving averages of the normalization; Default **0.1**
        relu (class, optional): Which ReLU to use; Default :class:`torch.nn.ReLU`

    Note:
        The bias term in all 3 :class:`~torch.nn.Conv2d` is disabled for this module.

    Note:
        This layer uses a residual connection for easier propagation of the gradient.
        Whether or not this residual connection is made, depends on the ``in_channels``, ``out_channels`` and ``stride`` arguments.
        If the input channels and output channels are equal and the stride is equal to 1, the residual connection is made.

    Example:
        >>> # In_channels is not equal to out_channels, so no residual connection
        >>> module = ln.network.layer.InvertedBottleneck(3, 32, 3, 1, 3)
        >>> print(module)
        InvertedBottleneck(3, 32, kernel_size=(3, 3), stride=(1, 1), expansion=3, ReLU(inplace=True))
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 10, 10])

        >>> # Stride is not 1, so no residual connection
        >>> module = ln.network.layer.InvertedBottleneck(32, 32, 3, 2, 3)
        >>> print(module)
        InvertedBottleneck(32, 32, kernel_size=(3, 3), stride=(2, 2), expansion=3, ReLU(inplace=True))
        >>> in_tensor = torch.rand(1, 32, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 5, 5])

        >>> # Residual connection is made
        >>> module = ln.network.layer.InvertedBottleneck(32, 32, 3, 1, 3)
        >>> print(module)
        InvertedBottleneck(32, 32, kernel_size=(3, 3), stride=(1, 1), expansion=3, ReLU(inplace=True), residual_connection)
        >>> in_tensor = torch.rand(1, 32, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 10, 10])
    """
    def __init__(self, in_channels, out_channels, kernel_size, stride, expansion, momentum=0.1, relu=lambda: nn.ReLU(inplace=True)):
        super().__init__()

        # Parameters
        self.expansion = expansion
        self.residual_connect = stride == 1 and in_channels == out_channels

        # Layer
        self.layers = nn.Sequential(
            nn.Conv2d(in_channels, in_channels*expansion, 1, 1, 0, bias=False),
            nn.BatchNorm2d(in_channels*expansion, momentum=momentum),
            relu(),

            nn.Conv2d(in_channels*expansion, in_channels*expansion, kernel_size, stride, 1, groups=in_channels*expansion, bias=False),
            nn.BatchNorm2d(in_channels*expansion, momentum=momentum),
            relu(),

            nn.Conv2d(in_channels*expansion, out_channels, 1, 1, 0, bias=False),
            nn.BatchNorm2d(out_channels, momentum=momentum),
        )

    def __repr__(self):
        residual = ', residual_connection' if self.residual_connect else ''
        s = '{name}({in_channels}, {out_channels}, kernel_size={kernel_size}, stride={stride}, expansion={expansion}, {relu}{residual})'
        return s.format(
            name=self.__class__.__name__,
            in_channels=self.layers[0].in_channels,
            out_channels=self.layers[6].out_channels,
            kernel_size=self.layers[3].kernel_size,
            stride=self.layers[3].stride,
            expansion=self.expansion,
            relu=self.layers[2],
            residual=residual,
        )

    def forward(self, x):
        if self.residual_connect:
            return x + self.layers(x)
        else:
            return self.layers(x)



#
#   Lightnet logger: Logging functionality used within the lightnet package
#   Copyright EAVISE
#
import os
import sys
import types
import logging
import copy
from enum import Enum

__all__ = ['logger']


# Formatter
class ColorCode(Enum):
    """ Color Codes """
    RESET = '\033[00m'
    BOLD = '\033[01m'

    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    WHITE = '\033[37m'
    GRAY = '\033[1;30m'


class ColoredFormatter(logging.Formatter):
    def __init__(self, msg, color=True, **kwargs):
        logging.Formatter.__init__(self, msg, **kwargs)
        self.color = color
        self.color_codes = {
            'CRITICAL': ColorCode.RED,
            'ERROR': ColorCode.RED,
            'TRAIN': ColorCode.BLUE,
            'TEST': ColorCode.BLUE,
            'DEPRECATED': ColorCode.YELLOW,
            'EXPERIMENTAL': ColorCode.YELLOW,
            'WARNING': ColorCode.YELLOW,
            'INFO': ColorCode.WHITE,
            'DEBUG': ColorCode.GRAY,
            'METADATA': ColorCode.GRAY,
        }

    def format(self, record):
        record = copy.copy(record)
        levelname = record.levelname
        name = record.name
        if self.color:
            color = self.color_codes[levelname] if levelname in self.color_codes else ''
            record.levelname = f'{ColorCode.BOLD.value}{color.value}{levelname:12}{ColorCode.RESET.value}'
        else:
            record.levelname = f'{levelname:12}'
        return logging.Formatter.format(self, record)

    def setColor(self, value):
        """ Enable or disable colored output for this handler """
        self.color = value


# Filter
class LevelFilter(logging.Filter):
    def __init__(self, levels, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.levels = levels

    def filter(self, record):
        if self.levels is None or record.levelname in self.levels:
            return True
        else:
            return False


# Logging levels
def log_once_function(lvl, msg_set):
    def log_once(self, message, *args, **kwargs):
        if not hasattr(self, msg_set):
            setattr(self, msg_set, set())

        msgs = getattr(self, msg_set)
        if self.isEnabledFor(lvl) and message not in msgs:
            msgs.add(message)
            self._log(lvl, message, args, **kwargs)

    return log_once


def log_function(lvl):
    def log(self, message, *args, **kwargs):
        if self.isEnabledFor(lvl):
            self._log(lvl, message, args, **kwargs)

    return log


def test(self, message, *args, **kwargs):
    if self.isEnabledFor(38):
        self._log(38, message, args, **kwargs)


def train(self, message, *args, **kwargs):
    if self.isEnabledFor(39):
        self._log(39, message, args, **kwargs)


# Metadata should usually not be printed to the console and is mainly used for logfiles
logging.addLevelName(1, 'METADATA')
logging.Logger.metadata = log_function(1)

# Experimental is a special warning mode, and thus it should be filtered the same (lower) as warning
logging.addLevelName(28, 'EXPERIMENTAL')
logging.Logger.experimental = log_once_function(28, 'experimental_msgs')

# Deprecated is a special warning mode, and thus it should be filtered the same (lower) as warning
logging.addLevelName(29, 'DEPRECATED')
logging.Logger.deprecated = log_once_function(29, 'deprecated_msgs')

# error_once function logs an error, but only once (useful in loops, but slows down code!)
logging.Logger.error_once = log_once_function(40, 'error_msgs')

# Test is a special info log, but with a much higher level. We only filter it with the highest log setting
logging.addLevelName(48, 'TEST')
logging.Logger.test = log_function(48)

# Train is a special info log, but with a much higher level. We only filter it with the highest log setting
logging.addLevelName(49, 'TRAIN')
logging.Logger.train = log_function(49)


# Console Handler
ch = logging.StreamHandler()
ch.setFormatter(ColoredFormatter('{levelname} {message}', style='{'))
if 'LN_LOGLVL' in os.environ:
    lvl = os.environ['LN_LOGLVL'].upper()
    try:
        ch.setLevel(int(lvl))
    except ValueError:
        ch.setLevel(lvl)

    if ch.level <= 10:
        ch.setFormatter(ColoredFormatter('{levelname} [{name}] {message}', style='{'))
else:
    ch.setLevel(logging.INFO)


# File Handler
def createFileHandler(self, filename, levels=None, filemode='a'):
    """ Create a file to write log messages of certaing levels """
    fh = logging.FileHandler(filename=filename, mode=filemode)
    fh.setLevel(logging.NOTSET)
    fh.addFilter(LevelFilter(levels))
    fh.setFormatter(logging.Formatter('{levelname} [{name}] {message}', style='{'))
    logger.addHandler(fh)
    return fh


# Logger
logger = logging.getLogger('lightnet')
logger.setLevel(1)
logger.addHandler(ch)
logger.setConsoleLevel = ch.setLevel
logger.setConsoleColor = ch.formatter.setColor
logger.setLogFile = types.MethodType(createFileHandler, logger)

# Disable color if ANSI not supported -> Code taken from django.core.management.color.supports_color
# Note that if you use the colorama plugin, you can reenable the colors
supported_platform = sys.platform != 'Pocket PC' and (sys.platform != 'win32' or 'ANSICON' in os.environ)
is_a_tty = hasattr(ch.stream, 'isatty') and ch.stream.isatty()
if not supported_platform or not is_a_tty:
    logger.setConsoleColor(False)



#
#   Darknet YOLOv2 with sensor fusion
#   Copyright EAVISE
#

import functools
from collections import OrderedDict, Iterable
import torch
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['YoloFusion']


class YoloFusion(lnn.module.Lightnet):
    """ Yolo v2 network that is able to fuse multiple image sensor data at a certain parameterizable point in the network :cite:`rgbd_fusion_v2`.

    Args:
        num_classes (Number, optional): Number of classes; Default **20**
        input_channels (int, optional): Number of input channels for the main subnetwork; Default **3**
        fusion_channels (int, optional): Number of input channels for the fusion subnetwork; Default **1**
        fuse_layer (int, optional): Number between 0-28, that controls at which layer to fuse both convolutional streams; Default **0**
        anchors (list, optional): 2D list with anchor values; Default **Yolo v2 anchors**

    Attributes:
        self.stride: Subsampling factor of the network (input_dim / output_dim)
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)

    Note:
        This network effectively supersedes the networks from :cite:`rgbd_fusion_v1`. |br|
        If you still want the old networks, you can take a look at a `previous version of lightnet`_.
        (Note that this is an older version of lightnet and thus some things might be broken in the latest version)

    .. _previous version of lightnet: https://gitlab.com/EAVISE/lightnet/blob/59baa61e429f63f80334dfff3ec2304d788ba1ad/lightnet/models/_network_yolo_fusion.py
    """
    stride = 32
    inner_stride = 32

    def __init__(self, num_classes, input_channels=3, fusion_channels=1, fuse_layer=0,
                 anchors=[(1.3221, 1.73145), (3.19275, 4.00944), (5.05587, 8.09892), (9.47112, 4.84053), (11.2364, 10.0071)]):
        super().__init__()
        if not isinstance(anchors, Iterable) and not isinstance(anchors[0], Iterable):
            raise TypeError('Anchors need to be a 2D list of numbers')
        if input_channels < 1 or fusion_channels < 1:
            raise ValueError('input_channels and fusion_channels need to be at least 1 [{input_channels}, {fusion_channels}]')

        # Parameters
        self.num_classes = num_classes
        self.anchors = anchors
        self.input_channels = input_channels
        self.fusion_channels = fusion_channels
        self.fuse_layer = fuse_layer
        self.fuse_seq = None

        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01

        # First layer
        if fuse_layer == 0:
            self.fuse_seq = 0
            self.layers = [
                nn.Sequential(OrderedDict([
                    ('fuse', nn.Conv2d(input_channels+fusion_channels, input_channels, 1, 1, 0, bias=False)),
                    ('1_convbatch', lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 1, 1, relu=relu, momentum=momentum)),
                ]))
            ]
        elif fuse_layer == 1:
            self.fuse_seq = 0
            self.layers = [
                nn.ModuleDict({
                    '1_convbatch_regular': lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 1, 1, relu=relu, momentum=momentum),
                    '1_convbatch_fusion': lnn.layer.Conv2dBatchReLU(fusion_channels, 32, 3, 1, 1, relu=relu, momentum=momentum),
                    'fuse': nn.Conv2d(32*2, 32, 1, 1, 0, bias=False),
                })
            ]
        else:
            self.layers = [
                nn.ModuleDict({
                    '1_convbatch_regular': lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 1, 1, relu=relu, momentum=momentum),
                    '1_convbatch_fusion': lnn.layer.Conv2dBatchReLU(fusion_channels, 32, 3, 1, 1, relu=relu, momentum=momentum),
                })
            ]

        # Main layers
        layer_list = [
            OrderedDict([
                ('2_max',           nn.MaxPool2d(2, 2)),
                ('3_convbatch',     lnn.layer.Conv2dBatchReLU(32, 64, 3, 1, 1, relu=relu, momentum=momentum)),
                ('4_max',           nn.MaxPool2d(2, 2)),
                ('5_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ('6_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                ('7_convbatch',     lnn.layer.Conv2dBatchReLU(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ('8_max',           nn.MaxPool2d(2, 2)),
                ('9_convbatch',     lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ('10_convbatch',    lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                ('11_convbatch',    lnn.layer.Conv2dBatchReLU(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ('12_max',          nn.MaxPool2d(2, 2)),
                ('13_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('14_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                ('15_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('16_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                ('17_convbatch',    lnn.layer.Conv2dBatchReLU(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
            ]),

            OrderedDict([
                ('18_max',          nn.MaxPool2d(2, 2)),
                ('19_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('20_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                ('21_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('22_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                ('23_convbatch',    lnn.layer.Conv2dBatchReLU(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('24_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('25_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
            ]),

            OrderedDict([
                ('26_convbatch',    lnn.layer.Conv2dBatchReLU((4*64)+1024, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('27_conv',         nn.Conv2d(1024, len(self.anchors)*(5+self.num_classes), 1, 1, 0)),
            ])
        ]
        i = 1
        for e, l in enumerate(layer_list, 1):
            if fuse_layer - i <= 0:
                fuse = None
            elif fuse_layer - i <= len(l):
                fuse = fuse_layer - i
                self.fuse_seq = e
            else:
                fuse = len(l) + 1
            i += len(l)

            self.layers.append(lnn.layer.Fusion(l, fuse))

        if self.fuse_seq is None:
            raise ValueError(f'Fuse_layer too high [{fuse_layer}/{sum([len(l) for l in layer_list])+1}]')

        # Passthrough layer
        if self.fuse_seq <= 2:
            self.layers.append(nn.Sequential(
                OrderedDict([
                    ('P1_convbatch',    lnn.layer.Conv2dBatchReLU(512, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('P2_reorg',        lnn.layer.Reorg(2)),
                ])
            ))
        else:
            self.layers.append(lnn.layer.Fusion(
                OrderedDict([
                    ('P1_convbatch',    lnn.layer.Conv2dBatchReLU(512, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                    ('P2_reorg',        lnn.layer.Reorg(2)),
                ]), 3
            ))

        self.layers = nn.ModuleList(self.layers)

    def forward(self, x):
        if x.size(1) != self.input_channels + self.fusion_channels:
            raise TypeError(f'This network requires {self.input_channels+self.fusion_channels} channel input images')

        # First layer
        if self.fuse_layer <= 0:
            x = self.layers[0](x)
        else:
            r = self.layers[0]['1_convbatch_regular'](x[:, :self.input_channels])
            f = self.layers[0]['1_convbatch_fusion'](x[:, self.input_channels:])
            x = torch.cat((r, f), 1)
            if 'fuse' in self.layers[0]:
                x = self.layers[0]['fuse'](x)

        # Sequence 1
        x = self.layers[1](x)

        # Passthrough
        if self.fuse_seq == 2:
            p = self.layers[4](x[:, :x.size(1)//2])
        else:
            p = self.layers[4](x)

        # Sequence 2
        x = self.layers[2](x)

        # Sequence 3
        if self.fuse_seq == 3:
            xs = x.size(1) // 2
            ps = p.size(1) // 2
            x = torch.cat((x[:, :xs], p[:, :ps], x[:, xs:], p[:, ps:]), 1)
        else:
            x = torch.cat((x, p), 1)
        x = self.layers[3](x)

        return x



#
#   Lightnet postprocessing brambox
#   Copyright EAVISE
#

import logging
import numpy as np
from ..util import BaseTransform
from ..._imports import pd, bb

__all__ = ['TensorToBrambox']
log = logging.getLogger(__name__)


class TensorToBrambox(BaseTransform):
    """ Converts a tensor to a list of brambox objects.

    Args:
        class_label_map (list, optional): List of class labels to transform the class id's in actual names; Default **None**

    Returns:
        pandas.DataFrame: brambox detection dataframe where the `image` column contains the batch number (int column from 1st tensor column).

    Note:
        Just like everything in PyTorch, this transform only works on batches of images.
        This means you need to wrap your tensor of detections in a list if you want to run this transform on a single image.

    Warning:
        If no `class_label_map` is given, this transform will simply convert the class id's to a string.
    """
    def __init__(self, class_label_map=None):
        super().__init__()
        self.class_label_map = class_label_map
        if self.class_label_map is None:
            log.warning('No class_label_map given. The indexes will be used as class_labels.')

    def forward(self, boxes):
        if boxes.numel() == 0:
            df = pd.DataFrame(columns=['image', 'class_label', 'id', 'x_top_left', 'y_top_left', 'width', 'height', 'confidence'])
            df.image = df.image.astype(int)
            df.class_label = df.class_label.astype(str)
            return df

        boxes = boxes.clone()

        # coords: width & height
        boxes[:, 3:5] -= boxes[:, 1:3]

        # Convert to brambox df
        boxes = boxes.cpu().detach().numpy()
        boxes = pd.DataFrame(boxes, columns=['image', 'x_top_left', 'y_top_left', 'width', 'height', 'confidence', 'class_label'])

        # Set column types
        boxes[['image', 'class_label']] = boxes[['image', 'class_label']].astype(int)
        boxes['id'] = np.nan
        boxes[['x_top_left', 'y_top_left', 'width', 'height', 'confidence']] = boxes[['x_top_left', 'y_top_left', 'width', 'height', 'confidence']].astype(float)

        if self.class_label_map is not None:
            boxes.class_label = boxes.class_label.map(dict((i, l) for i, l in enumerate(self.class_label_map)))
        else:
            boxes.class_label = boxes.class_label.astype(str)

        return boxes[['image', 'class_label', 'id', 'x_top_left', 'y_top_left', 'width', 'height', 'confidence']]



"""
Lightnet Network Module |br|
This module contains classes and functions to create deep neural network with pytorch_.
It is mostly targeted at Object detection networks, but contains the necessary tools to build any CNN type network.
"""

from . import layer
from . import loss
from . import module



#
#   Mobile Darknet19 model
#   Copyright EAVISE
#

import functools
from collections import OrderedDict
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['MobileDarknet19']


class MobileDarknet19(lnn.module.Darknet):
    """ Darknet19 implementation with depthwise separable convolutions.

    Args:
        num_classes (Number, optional): Number of classes; Default **1000**
        input_channels (Number, optional): Number of input channels; Default **3**

    Attributes:
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)
    """
    inner_stride = 32

    def __init__(self, num_classes, input_channels=3):
        super().__init__()

        # Parameters
        self.num_classes = num_classes
        self.input_channels = input_channels

        # Network
        relu = functools.partial(nn.LeakyReLU, 0.1, inplace=True)
        momentum = 0.01
        self.layers = nn.Sequential(
            # Base layers
            nn.Sequential(OrderedDict([
                ('1_convbatch',     lnn.layer.Conv2dBatchReLU(input_channels, 32, 3, 2, 1, relu=relu, momentum=momentum)),
                ('2_convdw',        lnn.layer.Conv2dDepthWise(32, 64, 3, 2, 1, relu=relu, momentum=momentum)),
                ('3_convdw',        lnn.layer.Conv2dDepthWise(64, 128, 3, 1, 1, relu=relu, momentum=momentum)),
                ('4_convbatch',     lnn.layer.Conv2dBatchReLU(128, 64, 1, 1, 0, relu=relu, momentum=momentum)),
                ('5_convdw',        lnn.layer.Conv2dDepthWise(64, 128, 3, 2, 1, relu=relu, momentum=momentum)),
                ('6_convdw',        lnn.layer.Conv2dDepthWise(128, 256, 3, 1, 1, relu=relu, momentum=momentum)),
                ('7_convbatch',     lnn.layer.Conv2dBatchReLU(256, 128, 1, 1, 0, relu=relu, momentum=momentum)),
                ('8_convdw',        lnn.layer.Conv2dDepthWise(128, 256, 3, 2, 1, relu=relu, momentum=momentum)),
                ('9_convdw',        lnn.layer.Conv2dDepthWise(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('10_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                ('11_convdw',       lnn.layer.Conv2dDepthWise(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('12_convbatch',    lnn.layer.Conv2dBatchReLU(512, 256, 1, 1, 0, relu=relu, momentum=momentum)),
                ('13_convdw',       lnn.layer.Conv2dDepthWise(256, 512, 3, 1, 1, relu=relu, momentum=momentum)),
                ('14_max',          nn.MaxPool2d(2, 2)),
                ('15_convdw',       lnn.layer.Conv2dDepthWise(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('16_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                ('17_convdw',       lnn.layer.Conv2dDepthWise(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
                ('18_convbatch',    lnn.layer.Conv2dBatchReLU(1024, 512, 1, 1, 0, relu=relu, momentum=momentum)),
                ('19_convdw',       lnn.layer.Conv2dDepthWise(512, 1024, 3, 1, 1, relu=relu, momentum=momentum)),
            ])),

            # Classification specific layers
            nn.Sequential(OrderedDict([
                ('20_conv',         nn.Conv2d(1024, num_classes, 1, 1, 0)),
                ('21_avgpool',      nn.AdaptiveAvgPool2d(1)),
                ('22_flatten',      lnn.layer.Flatten()),
            ])),
        )



import os
import subprocess
import setuptools as setup
from pkg_resources import get_distribution, DistributionNotFound


def get_dist(pkgname):
    try:
        return get_distribution(pkgname)
    except DistributionNotFound:
        return None

def find_packages():
    return ['lightnet'] + ['lightnet.'+p for p in setup.find_packages('lightnet')]

def get_version():
    with open('VERSION', 'r') as f:
        version = f.read().splitlines()[0]
    if version[-1] == 'a':  # Alpha dev-build
        try:
            cwd = os.path.dirname(os.path.abspath(__file__))
            sha = subprocess.check_output(['git', 'rev-parse', 'HEAD'], cwd=cwd).decode('ascii').strip()
            version += '-' + sha[:7]
        except Exception:
            pass

    with open('lightnet/version.py', 'w') as f:
        f.write('#\n')
        f.write('#   Lightnet version: Automatically generated version file\n')
        f.write('#\n\n')
        f.write(f'__version__ = "{version}"\n')

    return version


requirements = [
    'numpy',
    'torch',
    'torchvision',
    'brambox>=2',
]
pillow_req = 'pillow-simd' if get_dist('pillow-simd') is not None else 'pillow'
requirements.append(pillow_req)

setup.setup(
    name='lightnet',
    version=get_version(),
    author='EAVISE',
    description='Building blocks for recreating darknet networks in pytorch',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(),
    test_suite='test',
    install_requires=requirements,
)



"""
Lightnet Pruning Module |br|
This module contains classes and functions to prune convolutions in your networks,
in order to make them faster and/or better suited for your application.
"""

from . import dependency

from ._base import *
from ._multi import *
from ._l2 import *
from ._gm import *



import logging
import math
import torch
from ._base import *

__all__ = ['GMPruner']
log = logging.getLogger(__name__)


class GMPruner(Pruner):
    """ Geometric Median Pruning :cite:`gm_prune`.
    This pruner prunes channels which are closest to the geometric median of the channels in a convolution.

    The importance of a channel :math:`c_i` of convolution :math:`C` is computed as:

    .. math::
       Importance(c_i) = \\sum_{c_j \\in C} ||c_i - c_j||_p

    Args:
        model (torch.nn.Module): model to prune
        optimizer (torch.optim.Optimizer or None): Optimizer that is used when retraining the network (pass None if not retraining)
        input_dimensions (tuple): Input dimensions to the network
        p_norm (int, optional): Which norm to use when computing the importance; Default **2**
        manner ("soft" or "hard", optional): Whether to perform soft-pruning (replacing channel values with zero) or hard-pruning (deleting channels); Default **"hard"**
        get_parameters (function, optional): function that takes a model and returns the parameters for the optimizer; Default **model.parameters()**

    Warning:
        The percentage given is used on a per-layer basis for this pruning method.
        This means that if you give a percentage of 10%, we prune approximatively 10% of each prunable layer individually.

    Note:
        The percentage is approximative. |br|
        If the pruner tells to prune all filters in a convolution,
        we keep the most important and reduce the number of pruned filters by one.
    """
    def __init__(self, *args, p_norm=2, **kwargs):
        super().__init__(*args, **kwargs)
        self.p_norm = p_norm

    def prune(self, percentage, prune_manner):
        total_pruned = 0

        for key, dependency in self.dependencies.items():
            # Compute number of filters to prune
            num_pruned = math.floor(percentage * dependency.module.out_channels)
            if num_pruned >= dependency.module.out_channels:
                num_pruned = dependency.module.out_channels - 1
            if num_pruned <= 0:
                continue
            total_pruned += num_pruned

            # Get pairwise distance between all filters and sum them to get similarity score per filter
            w = dependency.module.weight.view(dependency.module.out_channels, -1)
            similarity = torch.cdist(w, w, p=self.p_norm).abs().sum(dim=0)

            # Prune selected filters
            indices = torch.argsort(similarity)[:num_pruned].tolist()
            prune_manner(dependency, indices)

        return total_pruned



#
#   MobileNet classification network
#   Copyright EAVISE
#

from collections import OrderedDict
import functools
import torch.nn as nn
import lightnet.network as lnn

__all__ = ['MobilenetV1']


class MobilenetV1(lnn.module.Lightnet):
    """ Mobilenet v1 classification network implementation :cite:`mobilenet_v1`.

    Args:
        num_classes (Number, optional): Number of classes; Default **1000**
        alpha (Number, optional): Number between [0-1] that controls the number of filters of the mobilenet convolutions; Default **1**
        input_channels (Number, optional): Number of input channels; Default **3**

    Attributes:
        self.inner_stride: Maximal internal subsampling factor of the network (input dimension should be a multiple of this)

    Note:
        The average pooling is implemented with an :class:`~torch.nn.AdaptiveAvgPool2d` layer. |br|
        For the base input dimension of 224x224, this is exactly the same as a 7x7 average pooling function,
        but the advantage of a adaptive average pooling is that this network can now handle multiple different input dimensions,
        as long as they are a multiple of the ``stride`` factor. |br|
        This is also how the implementation in `tensorflow <mobilenettf_>`_ (optionally) works.

    Warning:
        When changing the ``alpha`` value, you are changing the network architecture.
        This means you cannot use weights from this network with a different alpha value.

    .. _mobilenettf: https://github.com/tensorflow/models/blob/505f554c6417931c96b59516f14d1ad65df6dbc5/research/slim/nets/mobilenet_v1.py#L369-L378
    """
    inner_stride = 32

    def __init__(self, num_classes, alpha=1, input_channels=3):
        super().__init__()

        # Parameters
        self.num_classes = num_classes
        self.alpha = alpha
        self.input_channels = input_channels

        # Network
        relu = functools.partial(nn.ReLU6, inplace=True)
        self.layers = nn.Sequential(
            # Base layers
            nn.Sequential(OrderedDict([
                ('1_convbatch', lnn.layer.Conv2dBatchReLU(input_channels, int(alpha*32), 3, 2, 1, relu=relu)),
                ('2_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*32), int(alpha*64), 3, 1, 1, relu=relu)),
                ('3_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*64), int(alpha*128), 3, 2, 1, relu=relu)),
                ('4_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*128), int(alpha*128), 3, 1, 1, relu=relu)),
                ('5_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*128), int(alpha*256), 3, 2, 1, relu=relu)),
                ('6_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*256), int(alpha*256), 3, 1, 1, relu=relu)),
                ('7_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*256), int(alpha*512), 3, 2, 1, relu=relu)),
                ('8_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu)),
                ('9_convdw',    lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu)),
                ('10_convdw',   lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu)),
                ('11_convdw',   lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu)),
                ('12_convdw',   lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*512), 3, 1, 1, relu=relu)),
                ('13_convdw',   lnn.layer.Conv2dDepthWise(int(alpha*512), int(alpha*1024), 3, 2, 1, relu=relu)),
                ('14_convdw',   lnn.layer.Conv2dDepthWise(int(alpha*1024), int(alpha*1024), 3, 1, 1, relu=relu)),
            ])),

            # Classification specific layers
            nn.Sequential(OrderedDict([
                ('15_avgpool',  nn.AdaptiveAvgPool2d(1)),
                ('16_dropout',  nn.Dropout()),
                ('17_conv',     nn.Conv2d(int(alpha*1024), num_classes, 1, 1, 0)),
                ('18_flatten',  lnn.layer.Flatten()),
            ])),
        )



#
#   Cornernet pooling layers
#   Copyright EAVISE
#

from collections import OrderedDict
import torch.nn as nn
from ._darknet import Conv2dBatchReLU
from ._util import ParallelSum

__all__ = ['TopPool', 'BottomPool', 'LeftPool', 'RightPool', 'CornerPool']


class TopPool(nn.Module):
    """ Top pooling implementation :cite:`cornernet`. |br|
    This layer is not a traditional pooling layer in the sence that it does not modify the dimension of the input tensor.

    .. figure:: /.static/api/toppool.*
       :width: 100%
       :alt: TopPool module design

    Example:
        >>> module = ln.network.layer.TopPool()
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 3, 10, 10])
    """
    def forward(self, x):
        return x.flip(2).cummax(2)[0].flip(2)


class BottomPool(nn.Module):
    """ Bottom pooling implementation :cite:`cornernet`. |br|
    This layer is not a traditional pooling layer in the sence that it does not modify the dimension of the input tensor.

    .. figure:: /.static/api/bottompool.*
       :width: 100%
       :alt: BottomPool module design

    Example:
        >>> module = ln.network.layer.BottomPool()
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 3, 10, 10])
    """
    def forward(self, x):
        return x.cummax(2)[0]


class LeftPool(nn.Module):
    """ Left pooling implementation :cite:`cornernet`. |br|
    This layer is not a traditional pooling layer in the sence that it does not modify the dimension of the input tensor.

    .. figure:: /.static/api/leftpool.*
       :width: 100%
       :alt: LeftPool module design

    Example:
        >>> module = ln.network.layer.LeftPool()
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 3, 10, 10])
    """
    def forward(self, x):
        return x.flip(3).cummax(3)[0].flip(3)


class RightPool(nn.Module):
    """ Right pooling implementation :cite:`cornernet`. |br|
    This layer is not a traditional pooling layer in the sence that it does not modify the dimension of the input tensor.

    .. figure:: /.static/api/rightpool.*
       :width: 100%
       :alt: RightPool module design

    Example:
        >>> module = ln.network.layer.RightPool()
        >>> in_tensor = torch.rand(1, 3, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 3, 10, 10])
    """
    def forward(self, x):
        return x.cummax(3)[0]


class CornerPool(nn.Module):
    """ Cornerpooling module implementation :cite:`cornernet`.

    Args:
        channels (int): Number of input and output channels
        pool1 (nn.Module): First pooling module
        pool2 (nn.Module): Second pooling module
        inter_channels (int, optional): Intermediate channels; Default **128**
        momentum (number, optional): momentum of the moving averages of the normalization; Default **0.1**
        relu (class, optional): Which ReLU to use; Default :class:`torch.nn.ReLU`

    .. figure:: /.static/api/cornerpool.*
       :width: 100%
       :alt: CornerPool module design

    Example:
        >>> module = ln.network.layer.CornerPool(32, ln.network.layer.TopPool, ln.network.layer.BottomPool)
        >>> print(module)
        CornerPool(32, TopPool, BottomPool, inter_channels=128, ReLU(inplace=True))
        >>> in_tensor = torch.rand(1, 32, 10, 10)
        >>> out_tensor = module(in_tensor)
        >>> out_tensor.shape
        torch.Size([1, 32, 10, 10])

    Warning:
        Compared to the `official CornerNet implementation <cornernetImpl_>`_,
        this version of cornerpooling does not add the last 3x3 Conv2dBatchReLU module.

    .. _cornernetImpl: https://github.com/princeton-vl/CornerNet-Lite/blob/6a54505d830a9d6afe26e99f0864b5d06d0bbbaf/core/models/py_utils/utils.py#L187
    """
    def __init__(self, channels, pool1, pool2, inter_channels=128, momentum=0.1, relu=lambda: nn.ReLU(inplace=True)):
        super().__init__()
        self.layers = ParallelSum(OrderedDict([
            ('pool', ParallelSum(
                nn.Sequential(
                    Conv2dBatchReLU(channels, inter_channels, 3, 1, 1, momentum, relu),
                    pool1()
                ),
                nn.Sequential(
                    Conv2dBatchReLU(channels, inter_channels, 3, 1, 1, momentum, relu),
                    pool2()
                ),
                post=nn.Sequential(
                    nn.Conv2d(inter_channels, channels, 3, 1, 1, bias=False),
                    nn.BatchNorm2d(channels),
                )
            )),
            ('conv', nn.Sequential(
                nn.Conv2d(channels, channels, 1, 1, 0, bias=False),
                nn.BatchNorm2d(channels),
            )),
            ('post', relu())
        ]))

    def __repr__(self):
        s = '{name}({channels}, {pool1}, {pool2}, inter_channels={inter_channels}, {relu})'
        return s.format(
            name=self.__class__.__name__,
            channels=self.layers.pool[0][0].layers[0].in_channels,
            pool1=self.layers.pool[0][1].__class__.__name__,
            pool2=self.layers.pool[1][1].__class__.__name__,
            inter_channels=self.layers.pool[0][0].layers[0].out_channels,
            relu=self.layers.post,
        )

    def forward(self, x):
        return self.layers(x)



#
#   Lightnet optional dependencies
#   Copyright EAVISE
#
import logging

__all__ = ['onnx', 'onnx_numpy_helper']
log = logging.getLogger(__name__)

try:
    import onnx
    from onnx import numpy_helper as onnx_numpy_helper
except ModuleNotFoundError:
    log.warning('onnx is not installed and thus no pruning functionality will work')
    onnx = None
    onnx_numpy_helper = None



#
#   HyperParemeters container class
#   Copyright EAVISE
#
import copy
import logging
import importlib.util
from collections import Iterable
import torch

__all__ = ['HyperParameters']
log = logging.getLogger(__name__)


class HyperParameters:
    """ This class is a container for all your training hyperparameters.
    It allows to save the state of a training and reload it at a later stage.

    Args:
        **kwargs (dict, optional): Keywords arguments that will be set as attributes of the instance and serialized as well

    Attributes:
        self.batch: Number of batches processed; Gets initialized to **0**
        self.epoch: Number of epochs processed; Gets initialized to **0**
        self.*: All arguments passed to the initialization function can be accessed as attributes of this object

    Note:
        If you pass a ``kwarg`` that starts with an **_**,
        the parameter class will store it as a regular property without the leading **_**, but it will not serialize this variable.
        This allows you to store all parameters in this object, regardless of whether you want to serialize it.

        This also works when assigning new values after the object creation:

        >>> param = ln.engine.HyperParameters()
        >>> param._dummy = 666
        >>> print(param.dummy)
        666
    """
    __init_done = False
    __automatic = {'batch', 'epoch'}

    def __init__(self, **kwargs):
        for key in self.__automatic:
            setattr(self, key, 0)

        self.__no_serialize = []
        for key in kwargs:
            if key.startswith('_'):
                serialize = False
                val = kwargs[key]
                key = key[1:]
            else:
                serialize = True
                val = kwargs[key]

            if not hasattr(self, key):
                setattr(self, key, val)
                if not serialize:
                    self.__no_serialize.append(key)
            else:
                log.error(f'{key} attribute already exists as a HyperParameter and will not be overwritten.')

        self.__init_done = True

    def __setattr__(self, item, value):
        """ Store extra variables in this container class.
        This custom function allows to store objects after creation and mark whether are not you want to serialize them,
        by prefixing them with an underscore.
        """
        if item in self.__dict__ or not self.__init_done:
            super().__setattr__(item, value)
        elif item[0] == '_':
            if item[1:] in self.__dict__:
                raise AttributeError(f'{item} already stored in this object! Use {item[1:]} to access and modify it.')
            self.__no_serialize.append(item[1:])
            super().__setattr__(item[1:], value)
        else:
            super().__setattr__(item, value)

    def __repr__(self):
        """ Print all values stored in the object as repr.
        Objects that will not be serialized are marked with an asterisk.
        """
        s = f'{self.__class__.__name__}('
        for k in sorted(self.__dict__.keys()):
            if k.startswith('_HyperParameters__'):
                continue

            val = self.__dict__[k]
            valrepr = repr(val)
            if '\n' in valrepr:
                valrepr = valrepr.replace('\n', '\n    ')
            if k in self.__no_serialize:
                k += '*'

            s += f'\n  {k} = {valrepr}'

        return s + '\n)'

    def __str__(self):
        """ Print all values stored in the object as string.
        Objects that will not be serialized are marked with an asterisk.
        """
        s = f'{self.__class__.__name__}('
        for k in sorted(self.__dict__.keys()):
            if k.startswith('_HyperParameters__'):
                continue

            val = self.__dict__[k]
            valrepr = str(val)
            if '\n' in valrepr:
                valrepr = val.__name__ if hasattr(val, '__name__') else val.__class__.__name__
            if k in self.__no_serialize:
                k += '*'

            s += f'\n  {k} = {valrepr}'

        return s + '\n)'

    def __add__(self, other):
        """ Add 2 HyperParameters together. |br|
        This function first creates a deep copy of the first `self` argument
        and then loops through the items in the `other` argument and adds those parameters
        if they are not already available in the new HyperParameters object.

        Note:
            When adding HyperParameters objects together,
            we keep the automatic variables (epoch, batch) from the first object. |br|
            Optionally, you can reset these variables by calling the :func:`~lightnet.engine.HyperParameters.reset()` method.
        """
        if isinstance(other, HyperParameters):
            new = copy.deepcopy(self)
            return new.__iadd__(other)
        else:
            raise NotImplementedError('Can only add 2 Hyperparameters objects together')

    def __iadd__(self, other):
        # Small performance boost by not deepcopying self.
        if isinstance(other, HyperParameters):
            for key in other:
                if not hasattr(self, key):
                    nkey = f'_{key}' if key in other.__no_serialize else key
                    setattr(self, nkey, copy.deepcopy(getattr(other, key)))
                elif key not in HyperParameters.__automatic:
                    log.warn(f'"{key}" is available in both HyperParameters, keeping first')

            return self
        else:
            raise NotImplementedError('Can only add 2 Hyperparameters objects together')

    def keys(self):
        """ Returns the attributes of your HyperParameters object, similar to a python dictionary. """
        return sorted(k for k in self.__dict__ if not k.startswith('_HyperParameters_'))

    def values(self):
        """ Returns the attribute values of your HyperParameters object, similar to a python dictionary. """
        return (getattr(self, k) for k in self.keys())

    def items(self):
        """ Returns the attribute keys and values of your HyperParameters object, similar to a python dictionary. """
        return ((k, getattr(self, k)) for k in self.keys())

    def __iter__(self):
        """ Return an iterator of :func:`~lightnet.engine.HyperParameters.keys()`,
        so we can loop over this object like a python dictionary.
        """
        return iter(self.keys())

    @classmethod
    def from_file(cls, path, variable='params', **kwargs):
        """ Create a HyperParameter object from a dictionary in an external configuration file.
        This function will import a file by its path and extract a variable to use as HyperParameters.

        The main goal of this class is to enable *"python as a config"*. |br|
        This means that you can pass in a path to a python file,
        and the training code will automatically load the HyperParameters from this file.

        Args:
            path (str or path-like object): Path to the configuration python file
            variable (str, optional): Variable to extract from the configuration file; Default **'params'**
            **kwargs (dict, optional): Extra parameters that are passed to the extracted variable if it is a callable object

        Examples:
            .. doctest::
               :options: +SKIP

               >>> # config.py file
               >>> params = ln.engine.HyperParameters(
               ...     network = ln.models.YoloV2(20),
               ...     # _ means batch_size will not be serialized
               ...     _batch_size = 8,
               ...     lr = 0.001,
               ... )

               >>> # Main training/testing file
               >>> params = ln.engine.HyperParameters.from_file('config.py')
               >>> print(params)
               HyperParameters(
                   network = YoloV2
                   batch_size* = 8
                   lr = 0.001
               )

            By default, this function will look for a 'params' variable in your file,
            but you can change that by passing a different value to the ``variable`` argument.

            .. doctest::
               :options: +SKIP

               >>> # config.py file
               >>> my_custom_params = ln.engine.HyperParameters(...)

               >>> # Main training/testing file
               >>> params = ln.engine.HyperParameters.from_file('config.py', variable='my_custom_params')
               >>> print(params)
               HyperParameters(
                   ...
               )

            If you don't feel like importing lightnet in your config file, you can also return a dictionary,
            which will be used to construct a HyperParemeters object.

            .. doctest::
               :options: +SKIP

               >>> # config.py file
               >>> params = {
               ...     'a': 1,
               ...     'b': 2,
               ... }

               >>> # Main training/testing file
               >>> params = ln.engine.HyperParameters.from_file('config.py')
               >>> print(params)
               HyperParameters(
                   a = 1
                   b = 2
               )

            Finally, power users may want to be able to pass arguments to the config file! |br|
            Just make the 'params' argument callable in your config, and you can pass in keyword arguments.

            .. doctest::
               :options: +SKIP

               >>> # config.py file
               >>> def params(a, b):
               ...     # Either return a dict or an HP object
               ...     return {
               ...         'a': a,
               ...         'b': b,
               ...     }

               >>> # Main training/testing file
               >>> params = ln.engine.HyperParameters.from_file('config.py', a=666, b='value_B')
               >>> print(params)
               HyperParameters(
                   a = 666
                   b = value_B
               )

        Note:
            The extracted variable can be one of the following:

            - :class:`lightnet.engine.HyperParameters`: This object will simply be returned
            - ``dictionary``: The dictionary will be expanded as the parameters for initializing a new :class:`~lightnet.engine.HyperParameters` object
            - ``callable``: The object will be called with the optional kwargs and should return either a :class:`~lightnet.engine.HyperParameters` object or a ``dictionary``
        """
        try:
            spec = importlib.util.spec_from_file_location('lightnet.cfg', path)
            cfg = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(cfg)
        except AttributeError as err:
            raise ImportError(f'Failed to import the file [{path}]. Are you sure it is a valid python file?') from err

        try:
            params = getattr(cfg, variable)
        except AttributeError as err:
            raise AttributeError(f'Configuration variable [{variable}] not found in file [{path}]') from err

        if callable(params):
            params = params(**kwargs)

        if isinstance(params, cls):
            return params
        elif isinstance(params, dict):
            return cls(**params)
        else:
            raise TypeError(f'Unkown type for configuration variable {variable} [{type(params).__name__}]. This variable should be a dictionary or lightnet.engine.HyperParameters object.')

    def save(self, filename):
        """ Serialize all the hyperparameters to a pickle file. |br|
        The network, optimizers and schedulers objects are serialized using their ``state_dict()`` functions.

        Args:
            filename (str or path): File to store the hyperparameters

        Note:
            This function will first check if the existing attributes have a `state_dict()` function,
            in which case it will use this function to get the values needed to save.
        """
        state = {}

        for k, v in vars(self).items():
            if k not in self.__no_serialize:
                if hasattr(v, 'state_dict'):
                    state[k] = v.state_dict()
                else:
                    state[k] = v

        torch.save(state, filename)

    def load(self, filename, strict=True):
        """ Load the hyperparameters from a serialized pickle file.

        Note:
            This function will first check if the existing attributes have a `load_state_dict()` function,
            in which case it will use this function with the saved state to restore the values. |br|
            The `load_state_dict()` function will first be called with both the serialized value and the `strict` argument as a keyword argument.
            If that fails because of a TypeError, it is called with only the serialized value.
            This means that you will still get an error if the strict rule is not being followed,
            but functions that have a `load_state_dict()` function without `strict` argument can be loaded as well.
        """
        log.info(f'Loading state from file [{filename}]')
        state = torch.load(filename, 'cpu')

        for k, v in state.items():
            if hasattr(self, k):
                current = getattr(self, k)
                if hasattr(current, 'load_state_dict'):
                    try:
                        current.load_state_dict(v, strict=strict)
                    except TypeError:
                        current.load_state_dict(v)
                else:
                    setattr(self, k, v)
            else:
                setattr(self, k, v)

    def to(self, device):
        """ Cast the parameters from the network, optimizers and schedulers to a given device. |br|
        This function will go through all the class attributes and check if they have a `to()` function, which it will call with the device.

        Args:
            device (torch.device or string): Device to cast parameters

        Note:
            PyTorch optimizers and the ReduceLROnPlateau classes do not have a `to()` function implemented. |br|
            For these objects, this function will go through all their necessary attributes and cast the tensors to the right device.
        """
        for key, value in self.__dict__.items():
            if hasattr(value, 'to') and callable(value.to):
                value.to(device)
            elif isinstance(value, torch.optim.Optimizer):
                for param in value.state.values():
                    if isinstance(param, torch.Tensor):
                        param.data = param.data.to(device)
                        if param._grad is not None:
                            param._grad.data = param._grad.data.to(device)
                    elif isinstance(param, dict):
                        for subparam in param.values():
                            if isinstance(subparam, torch.Tensor):
                                subparam.data = subparam.data.to(device)
                                if subparam._grad is not None:
                                    subparam._grad.data = subparam._grad.data.to(device)
            elif isinstance(value, (torch.optim.lr_scheduler._LRScheduler, torch.optim.lr_scheduler.ReduceLROnPlateau)):
                for param in value.__dict__.values():
                    if isinstance(param, torch.Tensor):
                        param.data = param.data.to(device)
                        if param._grad is not None:
                            param._grad.data = param._grad.data.to(device)

    def reset(self):
        """ Resets automatic variables epoch and batch """
        for key in self.__automatic:
            setattr(self, key, 0)
